from django.contrib import messages
from django.http.request import HttpRequest
from django.template.response import TemplateResponse
from django.views.decorators.http import require_POST
from django.views.generic import FormView, TemplateView

from .forms import ContactForm, HTMXMessageForm
from .utils import is_htmx


@require_POST
def send_msg(request: HttpRequest):
    ctx = {"htmx_msg_form": HTMXMessageForm()}
    if is_htmx(request):
        if msg := request.POST.get("message"):
            level = int(request.POST.get("tag", 10))
            messages.add_message(request, level, msg)
    return TemplateResponse(request, "msg_form.html", ctx)


class HomePage(TemplateView):
    template_name = "home.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["htmx_msg_form"] = HTMXMessageForm()
        return context


class AboutPage(TemplateView):
    template_name = "about.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        messages.add_message(self.request, messages.DEBUG, "Test")
        messages.add_message(self.request, messages.INFO, "You are now informed")
        messages.add_message(self.request, messages.WARNING, "You warned")
        messages.add_message(self.request, messages.ERROR, "I'm an error!")
        messages.add_message(self.request, messages.SUCCESS, "I worked.")
        return context


class ContactFormView(FormView):
    template_name = "contact.html"
    form_class = ContactForm

    def form_invalid(self, form: ContactForm):
        return TemplateResponse(self.request, "contact.html", {"form": form})

    def form_valid(self, form: ContactForm):
        if form.is_valid():
            print("Success!")
            form = ContactForm()  # reset the form
        return TemplateResponse(self.request, "contact.html", {"form": form})



from bs4 import BeautifulSoup

from .filter_attrs import filter_attrs


def wrap_svg(html_markup: str, css: str | None = None, **kwargs) -> BeautifulSoup:
    """Supplement html fragment of `<svg>` icon with css classes and attributes, include parent/sibling `<span>`s when parameters dictate.

    The following kwargs: `pre_`, `post_`, and `parent_` args are respected.

    So `pre_text` + `pre_class` will add:

    ```html
     <!-- pre_ implies before the icon, with special rule for pre_text -->
    <span class='the-value-of-pre_class'>the-value-of-pre_text</span><svg></svg>
    ```

    `post_text` + `post_class` will add:

    ```html
    <!-- post_ implies after the icon, with special rule for post_text -->
    <svg></svg><span class='the-value-of-post_class'>the-value-of-post_text</span>
    ```

    `parent_class`  + `parent_title` will add:

    ```html
     <!-- parent_ implies a wrapper over the icon,
     'parent_text' will not have same effect.
     -->
    <span class='the-value-of-parent_class' title='the-value-of-parent_title'><svg></svg></span>
    ```

    Examples:
        >>> markup = '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 20 20" fill="currentColor" class="w-5 h-5"><path d="M6.28 5.22a.75.75 0 00-1.06 1.06L8.94 10l-3.72 3.72a.75.75 0 101.06 1.06L10 11.06l3.72 3.72a.75.75 0 101.06-1.06L11.06 10l3.72-3.72a.75.75 0 00-1.06-1.06L10 8.94 6.28 5.22z" /></svg>'
        >>> res = wrap_svg(html_markup=markup, pre_text="Close menu", pre_class="sr-only", aria_hidden="true")
        >>> len(res.contents) == 2
        True
        >>> res.contents[0]
        <span class="sr-only">Close menu</span>
        >>> res.contents[1].attrs == {'xmlns': 'http://www.w3.org/2000/svg', 'viewbox': '0 0 20 20', 'fill': 'currentColor', 'class': ['w-5', 'h-5'], 'aria-hidden': 'true'}
        True
        >>> parented = wrap_svg(html_markup=markup, parent_tag="button", pre_text="Close menu", pre_class="sr-only", aria_hidden="true")
        >>> elements = list(parented.children)
        >>> elements[0].name == 'button'
        True
        >>> list(elements[0].children)
        [<span class="sr-only">Close menu</span>, <svg aria-hidden="true" class="w-5 h-5" fill="currentColor" viewbox="0 0 20 20" xmlns="http://www.w3.org/2000/svg"><path d="M6.28 5.22a.75.75 0 00-1.06 1.06L8.94 10l-3.72 3.72a.75.75 0 101.06 1.06L10 11.06l3.72 3.72a.75.75 0 101.06-1.06L11.06 10l3.72-3.72a.75.75 0 00-1.06-1.06L10 8.94 6.28 5.22z"></path></svg>]

    Args:
        html_markup (str): The template that contains the `<svg>` tag converted into its html string format.
        css (str, optional): Previously defined CSS to add to the `<svg>` icon. Defaults to None.

    Returns:
        SafeString: Small HTML fragment visually representing an svg icon.
    """  # noqa: E501

    soup = BeautifulSoup(html_markup, "html.parser")
    icon = soup("svg")[0]
    if css:
        icon["class"] = css
    if aria_attrs := filter_attrs("aria", kwargs):
        for k, v in aria_attrs.items():
            icon[k] = v

    parent_tag = None
    if tagname := kwargs.pop("parent_tag", None):
        if tagname in ["button", "a", "span", "div"]:
            parent_tag = soup.new_tag(tagname)

    if parent_attrs := filter_attrs("parent", kwargs, unprefix=True):
        if not parent_tag:
            parent_tag = soup.new_tag("span")
        for k, v in parent_attrs.items():
            parent_tag[k] = v
    if parent_tag:
        icon.wrap(parent_tag)

    if left := kwargs.pop("pre_text", None):  # <span>, left of <svg>
        pre_span = soup.new_tag("span")
        pre_span.string = left
        if pre_attrs := filter_attrs("pre", kwargs, unprefix=True):
            for k, v in pre_attrs.items():
                pre_span[k] = v
        icon.insert_before(pre_span)
    elif right := kwargs.pop("post_text", None):  # <span>, right of <svg>
        post_span = soup.new_tag("span")
        post_span.string = right
        if post_attrs := filter_attrs("post", kwargs, unprefix=True):
            for k, v in post_attrs.items():
                post_span[k] = v
        icon.insert_after(post_span)
    return soup



from django.contrib.sites.shortcuts import get_current_site
from django.http.request import HttpRequest
from django.utils.html import format_html

from .fragments import register
from .helpers import strip_whitespace


@register.simple_tag
def og_title(text: str, request: HttpRequest | None = None):
    """Appends site name to end of title, if site name is available."""
    if request and (site := get_current_site(request)):
        if site.name not in text:
            text = f"{text} - {site.name}"
    return format_html(
        strip_whitespace("""
            <title>{text}</title>
            <meta property="og:title" content="{text}"/>
            <meta name="twitter:title" content="{text}"/>
            """),
        text=text,
    )


@register.simple_tag
def og_desc(text: str):
    """Appends description to open graph fields found in description template."""
    return format_html(
        strip_whitespace("""
            <meta name="description" content="{text}"/>
            <meta property="og:description" content="{text}"/>
            <meta name="twitter:description" content="{text}"/>
            """),
        text=text,
    )


@register.simple_tag
def og_img(url: str, alt: str):
    """Appends image details to open graph fields found in image template."""
    return format_html(
        strip_whitespace("""
            <meta property="og:image" content="{url}"/>
            <meta property="og:image:alt" content="{img_alt}"/>
            <meta name="twitter:image" content="{url}"/>
            <meta name="twitter:image:alt" content="{img_alt}"/>
            """),
        url=url,
        img_alt=alt,
    )



from pathlib import Path

from django import template
from django.conf import settings
from django.forms import BoundField
from django.http.request import HttpRequest
from django.template import Context, Template
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import SafeText, mark_safe
from markdown import markdown

from .utils import wrap_svg

register = template.Library()


@register.filter
def md(text: str, exts: list[str] = ["attr_list"]) -> SafeText:
    """Convert a text in markdown to its html equivalent using specific extensions."""
    return mark_safe(markdown(text, extensions=exts))


@register.simple_tag
def curr(lhs: str, reversible: str) -> SafeText:
    """Returns a string `aria-current` for use as an attribute when `lhs` path matches the
    `reversible` value that will be passed to the `django.urls.reverse()`.

    Args:
        lhs (str): lhs stands for lefthand side, should be first positional element in the tag
        reversible (str): The value will be passed to the `django.urls.reverse()` without args, kwargs

    Returns:
        SafeText: The text "aria-current=page" if a match occurs, otherwise ""
    """  # noqa: E501
    return mark_safe("aria-current=page" if lhs == reverse(reversible) else "")


@register.simple_tag
def nava(
    reversible: str,
    text: str | None = None,
    css: str | None = None,
    request: HttpRequest | None = None,
) -> SafeText:
    """HTML fragment: `<a>` tag for desktop/mobile navbar links.

    Checks if link represented by reverse name url `reversible` is the _current_ path via the request, if it is provided.

    In the latter case, on match of the `request.path` to the `reversible`, add `aria-current=true`

    Args:
        reversible (str): The value will be passed to the `django.urls.reverse()` function without keyword arguments
        text (str | None, optional): The text to incude within the anchor element, if any. Defaults to None.
        css (str | None, optional): If provided, will populate the `class` attribute of the anchor element. Defaults to None.
        request (HttpRequest | None, optional): The django http request object to ascertain the present path of the request. Defaults to None.

    Returns:
        SafeText: The output anchor tag
    """  # noqa: E501
    return format_html(
        "<a {aria} href='{url}' class='{css}'>{text}</a>",
        text=text or "",
        url=reverse(reversible),
        css=css,
        aria=curr(request.path, reversible) if request else "",
    )


@register.simple_tag
def icon(
    name: str,
    prefix: str = settings.FRAGMENTS.get("icons_prefix"),
    folder: Path = settings.FRAGMENTS.get("icons_path"),
    css: str | None = None,
    **kwargs,
) -> SafeText:
    """Make an `<svg>` fragment, using a file found within the `name` + `prefix` + `folder` path, and add the appropriate
    css classes and attributes, optionally including parent/sibling tags when parameters dictate.

    Args:
        name (str): The prefixless (prefix_) name of the `html` file containing an `<svg>` icon, presumes to be formatted and included in the proper folder previously.
        css (str, optional): Previously defined CSS to add to the `<svg>` icon. Defaults to None.
        prefix (str, optional): Source of the svg file; needs to be declared in `settings.py`.
        folder (str, optional): Where to find the template; needs to be declared in `settings.py`.
        **kwargs (dict): The following kwargs: `pre_`, `post_`, and `parent_` args are respected by `start_html_tag_helpers`

    Returns:
        SafeString: Small HTML fragment visually representing an svg icon but which may contain related tags.
    """  # noqa: E501
    path = folder / f"{prefix}_{name}.html"
    html = render_to_string(str(path))
    svg_with_kwargs = wrap_svg(html_markup=html, css=css, **kwargs)
    return mark_safe(str(svg_with_kwargs).strip())


@register.simple_tag
def hput(
    bound: BoundField,
    kls: str | None = "h",
    label_kls: str | None = None,
    validate: str | None = None,
) -> SafeText:
    """An encapsulation of an idiomatic BoundField with an optional ability for htmx inline
    field validation if a url is passed as `validate` argument.

    Args:
        bound (BoundField): The field to render
        kls (str | None, optional): If supplied will supply the containing div's class attribute. Defaults to "h".
        label_kls (str | None, optional): If supplied will supply the containing div's class attribute. Defaults to None.
        validate (str | None, optional): A url that should be the form's submit url for inline field validation using htmx. Defaults to None.

    Returns:
        SafeText: The html fragment consisting of a wrapped field
    """  # noqa: E501
    from .helpers import hx_enable_inline_validation

    if not isinstance(bound, BoundField):
        raise Exception(f"Improper field {bound=}")

    attrs = hx_enable_inline_validation(bound, validate) if validate else ""
    return mark_safe(
        Template("""
            {% load fragments %}
            {% whitespaceless %}
            <div {{attrs}}
                {% if bound.is_hidden %}hidden{% endif %}
                {% if bound.errors %}data-invalid=true{% endif %}
                class="{{ kls }}"
                data-widget="{{ bound.widget_type }}"
            >
                <label for="{{ bound.id_for_label }}"
                    {% if label_kls %}class="{{ label_kls }}"{% endif %}>
                    {{bound.label}}
                </label>
                {{ bound }}
                <small>{{ bound.help_text }}</small>
                {{ bound.errors }}
            </div>
            {% endwhitespaceless %}
            """)  # noqa: E501
        .render(
            context=Context(
                {
                    "bound": bound,
                    "label_kls": label_kls,
                    "kls": kls,
                    "attrs": attrs,
                },
            )
        )
        .strip()
    )



import re

from django import template
from django.forms import BoundField
from django.utils.functional import keep_lazy_text
from django.utils.safestring import SafeText, mark_safe

from .fragments import register


@register.simple_tag
def attrize(d: dict) -> SafeText:
    """Unpack a dictionary as attributes, useful for adding attributes to
    an existing tag."""
    return mark_safe(" ".join([f'{k}="{v}"' for k, v in d.items()]))


def hx_enable_inline_validation(bound: BoundField, url: str) -> str:
    """Given a `BoundField` and a `url`, implement inline field validation proposed by
    [hernantz](https://hernantz.github.io/inline-form-validation-with-django-and-htmx.html).
    It makes use of various tags from [htmx](https://htmx.org).

    The gist: form is prematurely submitted (because of `hx-post`) as an AJAX-request,
    the partial template response will `hx-swap` an `hx-target`. In other words, the
    old field is replaced by the same field... but now as a result of `form.is_valid()`.

    The response,if it contains errors, will include an error list for the field.

    Instead of rendering the entire partial response, the use of `hx-select` limits the
    replacement to a segment of the partial response.

    Args:
        bound (BoundField): A Django field with data previously submitted
        url (str): Where the form is submitted.
    Returns:
        str: A string of text attributes that can be added to a wrapping div of a `BoundField`
    """  # noqa: E501
    idx = f"hput_{bound.id_for_label}"
    return attrize(
        {
            "id": idx,
            "hx-select": f"#{idx}",
            "hx-post": url,
            "hx-trigger": "blur from:find input",
            "hx-target": f"#{idx}",
            "hx-swap": "outerHTML",
        }
    )


@keep_lazy_text
def strip_whitespace(value):
    """
    See https://stackoverflow.com/a/72942459 answer by [Will Gordon](https://stackoverflow.com/users/6758654/will-gordon)
    Return the given HTML with any newlines, duplicate whitespace, or trailing spaces are removed.
    """  # noqa: E501
    # Process duplicate whitespace occurrences or
    # *any* newline occurrences and reduce to a single space
    value = re.sub(r"\s{2,}|[\n]+", " ", str(value))
    # After processing all of the above,
    # any trailing spaces should also be removed
    # Trailing space examples:
    #   - <div >                    Matched by: \s(?=[<>"])
    #   - < div>                    Matched by: (?<=[<>])\s
    #   - <div class="block ">      Matched by: \s(?=[<>"])
    #   - <div class=" block">      Matched by: (?<==\")\s
    #   - <span> text               Matched by: (?<=[<>])\s
    #   - text </span>              Matched by: \s(?=[<>"])
    value = re.sub(r'\s(?=[<>"])|(?<==\")\s|(?<=[<>])\s', "", str(value))
    return value


class WhitespacelessNode(template.Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        return strip_whitespace(self.nodelist.render(context).strip())


@register.tag
def whitespaceless(parser, token):
    """Remove whitespace within HTML tags, including tab, newline and extra space characters. Warning: this affects all text within the `whitespaceless` command without prejudice. Use with caution."""  # noqa: E501
    nodelist = parser.parse(("endwhitespaceless",))
    parser.delete_first_token()
    return WhitespacelessNode(nodelist)



from django import forms
from django.db import models
from django.utils.translation import gettext_lazy as _


class HTMXMessageForm(forms.Form):
    template_name = "test_snippet.html"

    class MessageTag(models.IntegerChoices):
        DEBUG = 10
        INFO = 20
        SUCCESS = 25
        WARNING = 30
        ERROR = 40

    message = forms.CharField(label="Message", widget=forms.Textarea(attrs={"rows": 3}))
    tag = forms.TypedChoiceField(
        label="Tag",
        choices=MessageTag.choices,
        initial=MessageTag.DEBUG,
        help_text="This is a message tag",
    )


class ContactForm(forms.Form):
    class Category(models.TextChoices):
        YES = "yes", _("Yes, I approve")
        NO = "no", _("No, I do not")
        MAYBE = "maybe", _("I'll think about it")
        MEH = "meh", _("Won't even think")
        IGNORE = "ignore", _("No reply")

    template_name = "test_snippet.html"
    category = forms.TypedChoiceField(
        label="Thoughts",
        choices=Category.choices,
        initial=Category.NO,
        help_text="This is a category!",
    )
    email = forms.EmailField(label="Email", help_text="Testable form")
    message = forms.CharField(label="Message", widget=forms.Textarea(attrs={"rows": 3}))
